import os
import uuid
from enum import Enum


class user:
    uuid = str(uuid.uuid4())[:8]


class ipinfo:
    url = "https://ipinfo.io/json"


class geocoder:
    url = "https://nominatim.openstreetmap.org"


class renopsapi:
    # Test endpoint
    # renewable_potential = "http://127.0.0.1:8000/v1/forecast/renewable_potential"
    renewable_potential = (
        "https://renops-api-tango.xlab.si/v1/forecast/renewable_potential"
    )
    price = "https://renops-api-tango.xlab.si/v1/forecast/day_ahead_prices"
    carbon_emissions = "https://renops-api-tango.xlab.si/v1/forecast/carbon_emissions"

    key = os.getenv("RENOPSAPI_KEY")
    if key is None:
        msg = "RENOPSAPI_KEY environment variable is not set."
        msg += " Export it with your API key to remove this error."
        raise ValueError(msg)
    max_retries = 5
    secs_between_retries = 15


class runtime:
    verbose = False
    sleep_seconds = 10
    min_savings_perc = 5  # If savings too small, we execute now

    @classmethod
    def set_verbose(cls, value):
        cls.verbose = value


class OptimisationType(Enum):
    renewable_potential = "renewable_potential"
    price = "price"
    carbon_emissions = "carbon_emissions"



import unittest

from renops.geoshifter import GeoShift
from renops.scheduler import Scheduler


class SchedulerTestLocationOutsideEU(unittest.TestCase):
    """Test rp on set location outside EU"""

    def setUp(self):
        # Create a Scheduler instance with sample parameters
        self.scheduler = Scheduler(
            deadline=1,
            runtime=1,
            location="NY",
            action=lambda: print("Executing action"),
            verbose=True,
            optimise_type="renewable_potential",
            argument=(1, 2, 3),
            kwargs={"key": "value"},
        )
        self.data = self.scheduler.get_data()

    def test_get_data(self):
        # Test the get_data method
        self.assertFalse(self.data.isna().any().any())

    def test_preprocess_data(self):
        # Test the _preprocess_data method
        preprocessed_data = self.scheduler._preprocess_data(self.data)
        self.assertIsNotNone(preprocessed_data)
        # Add more assertions based on the expected behavior of the method


class SchedulerTestAutoLocation(unittest.TestCase):
    """Test rp on with auto location"""

    def setUp(self):
        # Create a Scheduler instance with sample parameters
        self.scheduler = Scheduler(
            deadline=1,
            runtime=1,
            location="auto",
            action=lambda: print("Executing action"),
            verbose=True,
            optimise_type="renewable_potential",
            argument=(1, 2, 3),
            kwargs={"key": "value"},
        )
        self.data = self.scheduler.get_data()

    def test_get_data(self):
        # Test the get_data method
        self.assertFalse(self.data.isna().any().any())

    def test_preprocess_data(self):
        # Test the _preprocess_data method
        preprocessed_data = self.scheduler._preprocess_data(self.data)
        self.assertIsNotNone(preprocessed_data)
        # Add more assertions based on the expected behavior of the method


class SchedulerTestPriceOptimisation(unittest.TestCase):
    """Test price optimisation"""

    def setUp(self):
        # Create a Scheduler instance with sample parameters
        self.scheduler = Scheduler(
            deadline=1,
            runtime=1,
            location="Berlin",
            action=lambda: print("Executing action"),
            verbose=True,
            optimise_type="price",
            argument=(1, 2, 3),
            kwargs={"key": "value"},
        )
        self.data = self.scheduler.get_data()

    def test_get_data(self):
        # Test the get_data method
        self.assertFalse(self.data.isna().any().any())

    def test_preprocess_data(self):
        # Test the _preprocess_data method
        preprocessed_data = self.scheduler._preprocess_data(self.data)
        self.assertIsNotNone(preprocessed_data)
        # Add more assertions based on the expected behavior of the method


class SchedulerTestCarbonEmissions(unittest.TestCase):
    """Test carbon emissions optimisation"""

    def setUp(self):
        # Create a Scheduler instance with sample parameters
        self.scheduler = Scheduler(
            deadline=1,
            runtime=1,
            location="Ljubljana",
            action=lambda: print("Executing action"),
            verbose=True,
            optimise_type="carbon_emissions",
            argument=(1, 2, 3),
            kwargs={"key": "value"},
        )
        self.data = self.scheduler.get_data()

    def test_get_data(self):
        # Test the get_data method
        self.assertFalse(self.data.isna().any().any())

    def test_preprocess_data(self):
        # Test the _preprocess_data method
        preprocessed_data = self.scheduler._preprocess_data(self.data)
        self.assertIsNotNone(preprocessed_data)
        # Add more assertions based on the expected behavior of the method


class GeoShiftTestsPotential(unittest.TestCase):
    def setUp(self):
        # Create a GeoShift instance with sample parameters
        self.locations = {
            "Location1": {"location": "Berlin", "cmd": "echo 'Hello, Berlin!'"},
            "Location2": {"location": "Paris", "cmd": "echo 'Bonjour, Paris!'"},
        }
        self.geoshifter = GeoShift(
            locations=self.locations, optimise_type="renewable_potential", verbose=True
        )

    def test_check_subkeys(self):
        # Test the check_subkeys method
        self.assertTrue(self.geoshifter.check_subkeys(self.locations))
        # Add more assertions based on the expected behavior of the method

    def test_shift(self):
        # Test the shift method
        stdout, stderr = self.geoshifter.shift()
        self.assertIsNotNone(stdout)
        self.assertEqual(stderr, "")
        # Add more assertions based on the expected behavior of the method


class GeoShiftTestsPrice(unittest.TestCase):
    def setUp(self):
        # Create a GeoShift instance with sample parameters
        self.locations = {
            "Location1": {"location": "Berlin", "cmd": "echo 'Hello, Berlin!'"},
            "Location2": {"location": "Paris", "cmd": "echo 'Bonjour, Paris!'"},
        }
        self.geoshifter = GeoShift(
            locations=self.locations, optimise_type="price", verbose=True
        )

    def test_check_subkeys(self):
        # Test the check_subkeys method
        self.assertTrue(self.geoshifter.check_subkeys(self.locations))
        # Add more assertions based on the expected behavior of the method

    def test_shift(self):
        # Test the shift method
        stdout, stderr = self.geoshifter.shift()
        self.assertIsNotNone(stdout)
        self.assertEqual(stderr, "")
        # Add more assertions based on the expected behavior of the method


if __name__ == "__main__":
    unittest.main()



import time

from loguru import logger

import renops.config as conf
from renops.datafetcher import DataFetcher
from renops.utils import execute_linux_command, get_closest_metric


class GeoShift:
    def __init__(
        self,
        locations: dict,
        optimise_type: str = "renewable_potential",
        verbose: bool = False,
    ):
        self.locations = locations
        self.verbose = verbose
        try:
            self.optimise_type = conf.OptimisationType[optimise_type]
        except KeyError:
            raise ValueError(
                f"Invalid option '{optimise_type}', must be one of {[e.value for e in conf.OptimisationType]}."
            )  # noqa

        self.optimise_type = optimise_type

    def check_subkeys(self, dictionary: dict) -> bool:
        for key, value in dictionary.items():
            if "location" not in value or "cmd" not in value:
                return False
        return True

    def shift(self):
        if not self.check_subkeys(self.locations):
            raise ValueError(
                'Each dictionary in the input dictionary must have a "location" and "cmd" key.'
            )

        metrics = {}
        for key, value in self.locations.items():
            current_epoch = int(time.time())
            forecast = DataFetcher(value["location"]).fetch(
                optimise_type=self.optimise_type
            )
            current_metric = get_closest_metric(forecast, current_epoch)
            if self.verbose:
                logger.info(
                    f"Current metric for {key} in {value['location']} is: {current_metric:.2f}"
                )
            metrics[key] = current_metric

        if self.optimise_type == "renewable_potential":
            best_location = max(metrics, key=metrics.get)
        else:
            best_location = min(metrics, key=metrics.get)

        ep = self.locations[best_location]

        logger.info(f'Found optimal location: {best_location}, {ep["location"]}!')
        logger.info(f"... Running specified command: {ep['cmd']}!")

        stdout, stderr = execute_linux_command(ep["cmd"])
        logger.info(f"stdout: {stdout}")
        logger.info(f"stderr: {stderr}")

        return stdout, stderr



import sys
import time
from typing import Dict, List, Union

import pandas as pd
import requests
from loguru import logger

import renops.config as conf
from renops.geolocation import GeoLocation


class DataFetcher:
    def __init__(self, location: Union[str, Dict[str, float]] = None):
        """
        Initialize the DataFetcher class.
        Args:
            url (str): The URL of the server.
            location (Union[str, Dict[str, float]]): The location, either as a string (city name) or as coordinates
                (latitude and longitude).
        """
        self.params = GeoLocation(location).params
        self._check_coordinates()

    def _check_coordinates(self):
        if self.params["lat"] is None:
            raise RuntimeError("Failed to obtain coordinates from Geocoder!")

    def filter_dict(self, in_dict: Dict, keys_to_keep: List) -> Dict:
        """
        Return dictionary with keys we want to keep
        """
        return {key: in_dict[key] for key in keys_to_keep}

    def _request_data(self, url: str) -> Dict:
        response = requests.get(
            url, params=self.params, headers={"key": conf.renopsapi.key}
        )
        response.raise_for_status()  # Raises an exception for 4xx or 5xx status codes

        return response.json()

    def _preporcess_data(self, response: Dict) -> pd.DataFrame:
        # Define needed keys for calculating renewable potential
        keys_to_keep = ["timestamps_hourly", "metric"]

        # Keep just the keys we need
        data = self.filter_dict(response, keys_to_keep)

        # Convert to DataFrame
        data = pd.DataFrame(data)

        # Get timestamp units
        unit = response["units"]["timestamps_hourly"]

        # Convert epoch strings to epoch ints
        data["epoch"] = data["timestamps_hourly"].astype(int)

        # Convert to datetime
        data["Date"] = pd.to_datetime(data["epoch"], unit=unit)
        data.set_index("Date", inplace=True)

        return data

    def fetch_renewable_potential(self) -> Dict:
        """
        Fetches the renewable potential data for the next 5 days.
        Returns:
            dict or None: The fetched data as a JSON object, or None if an error occurred.
        """
        # Get reponse
        response = self._request_data(conf.renopsapi.renewable_potential)

        # Rename columns
        response["metric"] = response["renewable_potential_forecast_hourly"]

        return response

    def fetch_prices(self) -> Dict:
        """
        Fetches day ahead prices.
        Returns:
            dict or None: The fetched data as a JSON object, or None if an error occurred.
        """
        # Get reponse
        response = self._request_data(conf.renopsapi.price)
        # Rename columns
        response["metric"] = response["day_ahead_prices"]

        return response

    def fetch_emissions(self) -> Dict:
        """Fetches carbon emissions from renops api

        Returns:
            Dict or none: Json if ok, None if an error occured
        """
        # Get response
        response = self._request_data(conf.renopsapi.carbon_emissions)
        # Rename column
        response["metric"] = response["carbon_emissions"]
        return response

    def fetch(self, optimise_type: str):
        max_retries = conf.renopsapi.max_retries

        for attempt in range(max_retries):

            try:
                if optimise_type == "price":
                    response = self.fetch_prices()
                elif optimise_type == "carbon_emissions":
                    response = self.fetch_emissions()
                elif optimise_type == "renewable_potential":
                    response = self.fetch_renewable_potential()
                else:
                    logger.error("optimise type not valid!")
                    sys.exit(1)

                return self._preporcess_data(response)

            except requests.exceptions.RequestException as e:
                logger.error(str(e))

                if "502" in str(e):
                    if attempt < max_retries - 1:
                        logger.error(
                            f"Requested endpoint is down {e}, waiting {conf.renopsapi.secs_between_retries} seconds before retry {attempt+1} / {max_retries} ..."  # noqa
                        )  # noqa
                        time.sleep(
                            conf.renopsapi.secs_between_retries
                        )  # Wait 30 seconds before retrying
                    else:
                        raise RuntimeError(
                            f"Failed to retrieve data after {max_retries} attempts"
                        ) from e
                if "422" in str(e):
                    logger.error(
                        "Could not map a bidding zone to given coordinate. "
                        "Check (https://www.entsoe.eu/network_codes/bzr/) for more details."
                    )
                    sys.exit(1)



import subprocess
import time
from datetime import datetime
from typing import Callable, Tuple, Union

from loguru import logger

import renops.config as conf
from renops.datafetcher import DataFetcher


def parse_time(time_string):
    return datetime.strptime(time_string, "%Y-%m-%d %H:%M:%S")


def wait_until(target_time):
    while int(time.time()) < target_time:
        time.sleep(conf.runtime.sleep_seconds)  # Sleep for a bit to not hog the CPU


def execute_script(script_path):
    subprocess.run(["python3", script_path])


def hour_to_second(hour: int) -> int:
    "Converts hours to seconds"
    return hour * 3600


def convert_seconds_to_hour(seconds: int) -> int:
    "Converts seconds to hour with no residual"
    return int(seconds // 3600)


def convert_seconds_to_minutes(seconds: int) -> int:
    "Converts seconds to minutes with no residual"
    return int((seconds % 3600) // 60)


def to_datetime(epoch):
    return datetime.fromtimestamp(epoch).strftime("%Y-%d-%m %H:%M:%S")


class Scheduler:
    """
    Scheduler to optmise scheduling of energy intensive tasks
    Args:
        deadline (int): The deadline for the scheduled task.
        runtime (int): The runtime required for the task.
        location (str): The location where the task will be executed.
        action (Callable): The function to be executed.
        optimise_type (str, optional): The type of optimization optons are "price", "carbon_emmisions",
                                       "renewable_potential". Defaults to "renewable_potential".
        optimise_price (bool, optional): Deprecated. Use 'optimise_type' instead. Defaults to False.
        verbose (bool, optional): Flag to enable verbose output. Defaults to False.
        argument (Tuple[Union[int, str], ...], optional): Arguments for the action function. Defaults to ().
        kwargs (Union[dict, None], optional): Keyword arguments for the action function. Defaults to {}.

    Raises:
        DeprecationWarning: If 'optimise_price' is used.
        ValueError: If an invalid 'optimise_type' is provided.

    Attributes:
        deadline (int): The deadline for the scheduled task.
        runtime (int): The runtime required for the task.
        location (str): The location where the task will be executed.
        v (bool): Verbose flag.
        optimise_type (str): The type of optimization.
        action (Callable): The function to be executed.
        argument (Tuple[Union[int, str], ...]): Arguments for the action function.
        kwargs (dict): Keyword arguments for the action function.
    """

    def __init__(
        self,
        deadline: int,
        runtime: int,
        location: str,
        action: Callable,
        optimise_type: str = "renewable_potential",
        optimise_price: bool = False,
        verbose: bool = False,
        argument: Tuple[Union[int, str], ...] = (),
        kwargs: Union[dict, None] = {},
    ) -> None:
        if optimise_price:
            optimise_type = "price"
            raise DeprecationWarning(
                "'optimise_price' is deprecated and will be removed in future versions. Use 'optimise_type' instead."
            )
        self.deadline = deadline
        self.runtime = runtime
        self.location = location
        self.v = verbose
        # Allow optimise_type to be a string
        # Convert string optimise_type to OptimisationType enum
        try:
            self.optimise_type = conf.OptimisationType[optimise_type]
        except KeyError:
            raise ValueError(
                f"Invalid option '{optimise_type}', must be one of {[e.value for e in conf.OptimisationType]}."
            )  # noqa

        self.optimise_type = optimise_type
        self.action = action
        self.argument = argument
        self.kwargs = kwargs if kwargs is not None else {}

    def get_data(self):
        fetcher = DataFetcher(location=self.location)
        return fetcher.fetch(self.optimise_type)

    def _preprocess_data(self, data):

        # Resample to 2H buckets
        res = data.resample("2h").agg(
            {
                "metric": "mean",
                "epoch": "first",
                "timestamps_hourly": "first",
            }
        )

        # Sort to minimise renewable potential
        res = res.set_index("epoch")

        # In case of rewewable potential we maximise, in case of price and
        # emissions we minimise
        ascending = False if self.optimise_type == "renewable_potential" else True

        res = res.sort_values(by=["metric"], ascending=ascending)

        return res

    def _extract_epochs(self):
        self.current_epoch = int(time.time())
        self.deadline_epoch = self.current_epoch + hour_to_second(self.deadline)
        self.start_execution_epoch = self.deadline_epoch - hour_to_second(self.runtime)

        return None

    def _filter_samples(self, res):
        filtered_res = res[
            (res.index >= self.current_epoch)
            & (res.index <= self.start_execution_epoch)
        ]
        filtered_res = filtered_res.loc[res.metric != 0]
        return filtered_res

    def _get_current_renewables(self, data):
        renewables_now = data[data.epoch >= self.current_epoch]
        renewables_now = renewables_now.metric.values[0].round(2)
        return renewables_now

    def _update_global_config(self):
        conf.runtime.set_verbose(self.v)

    def _calculate_estimated_improvement(self, val_now: float, val_min: float) -> float:
        """Returns estimated precent reducing given optmisation type
        The lower the better, i.e. -5 would save 5 %.

        Args:
            data (pd.DataFrame): dataframe populated with predictions
        """
        diff = val_min - val_now
        diff_perc = (diff / val_now) * 100

        if self.optimise_type in ["price", "carbon_emissions"]:
            diff_perc *= (
                -1
            )  # In case of price and carbon we minimise, in case of renewable we maxismise

        return diff_perc

    def _print_info_for_instant_execution(self):
        if self.optimise_type == "price":
            logger.info(f"Current energy price is: {self.renewables_now} EUR/MWh")
        elif self.optimise_type == "carbon_emissions":
            logger.info(
                f"Current carbon emissions are: {self.renewables_now} gCO2eq/kWh"
            )
        elif self.optimise_type == "renewable_potential":
            logger.info(f"Current renewable potential is: {self.renewables_now}")

    def run(self):
        self._update_global_config()
        data = self.get_data()
        res = self._preprocess_data(data)
        self._extract_epochs()  # extract deadilnes runtimes etc TODO
        filtered_res = self._filter_samples(res)
        self.renewables_now = self._get_current_renewables(data)
        if self.v:
            logger.info(
                f"Task has to be finished by: {to_datetime(self.deadline_epoch)}"
            )
        if len(filtered_res) <= 1:
            filtered_res[self.current_epoch] = self.renewables_now
            optimal_time = self.current_epoch

            if self.v:
                logger.info("No renewable window whitin a given deadline!")
                self._print_info_for_instant_execution()

        else:
            optimal_time = filtered_res.index[0]
            diff_seconds = optimal_time - self.current_epoch
            current_val = self._get_current_renewables(data)
            minimal_val = filtered_res.metric.values[0].round(2)
            diff_perc = self._calculate_estimated_improvement(current_val, minimal_val)

            if diff_perc < conf.runtime.min_savings_perc:
                logger.info(
                    f"Estimated savings ({diff_perc:.2f}) % are too low, executing now!"
                )
                if self.v:
                    self._print_info_for_instant_execution()
            else:
                if self.v:
                    logger.info(
                        f"Found optimal time between {to_datetime(filtered_res.index[0])} and {to_datetime(filtered_res.index[0] + hour_to_second(self.runtime))}",  # noqa
                    )

                    if self.optimise_type == "price":
                        msg = f"Energy price at that time is: {minimal_val} EUR/MWh"

                    elif self.optimise_type == "carbon_emissions":
                        msg = f"Carbon emissions at that time are: {minimal_val} gCO2eq/kWh"

                    elif self.optimise_type == "renewable_potential":
                        msg = f"Renewable potential at that time is: {minimal_val}"

                    logger.info(
                        msg + f", where esitmated savings are {diff_perc:.2f} %"
                    )
                    logger.info(
                        f"Waiting for"
                        f" {convert_seconds_to_hour(diff_seconds)} h"
                        f" {convert_seconds_to_minutes(diff_seconds)} min"
                        f"..."
                    )

        wait_until(optimal_time)

        if self.v:
            logger.info(f"Executing action now at {datetime.now()}")
        if self.v:
            logger.info("----------------------------------------------------")
        self.action(*self.argument, **self.kwargs)



from typing import Any, Dict, Tuple, Union

import requests
from loguru import logger

import renops.config as conf


class GeoLocation:
    def __init__(self, location: Union[str, Dict[str, float]] = None):
        self.params = self._get_location_params(location)

    def _get_location_params(
        self, location: Union[None, str, Dict[str, float]]
    ) -> Dict[str, Any]:
        """
        Returns the parameters based on the provided location (city name or coordinates).
        Args:
            location (Union[str, Dict[str, float]]): The location, either as a string (city name) or as coordinates
                (latitude and longitude). In case None, automatic location detection is executed.
        Returns:
            dict: The parameters for the request.
        """
        auto_synonyms = ["auto", "a", "automatic"]  # Define synomims for word automatic
        if isinstance(location, str) and location not in auto_synonyms:
            # When location is defined as a word
            lat, lon = self._geocode_location(location)
            if conf.runtime.verbose:
                logger.info(f"Location specified: {location}, lat: {lat} lon: {lon}")
        elif location in auto_synonyms:
            # When location is set to auto
            loc = self._get_location()
            if conf.runtime.verbose:
                logger.warning(
                    f'Location is set to auto, IP will be used to detect location! found: {loc["city"]},'
                    f' {loc["country"]}'
                )
            lat, lon = loc["loc"].split(",")
        # elif isinstance(location, dict) and "lat" in location and "lon" in location:
        #    lat, lon = location["lat"], location["lon"]
        #    logger.info(f"Location specified: {location}")
        else:
            raise ValueError("Invalid location format")
        return {"lat": lat, "lon": lon}

    def _geocode_location(self, location: str) -> Tuple[float, float]:
        """
        Geocodes the provided location (city name) to obtain latitude and longitude.
        Args:
            location (str): The location (city name) to geocode.
        Returns:
            tuple: The latitude and longitude values.
        """
        headers = {
            "Referer": "https://renops-api-tango.xlab.si",
            "User-Agent": f"renops-scheduler-user-{conf.user.uuid}",
        }
        response = requests.get(
            f"{conf.geocoder.url}/search?q={location}&format=json", headers=headers
        )
        # Check if the API request was successful
        if response.status_code == 200:
            data = response.json()
            # Check if the settlement was found
            if len(data) > 0:
                # Extract the latitude and longitude from the API response
                lat = data[0]["lat"]
                lng = data[0]["lon"]
                return lat, lng
            else:
                logger.info(f'Settlement "{location}" not found.')
                return None, None
        else:
            logger.error(
                f"API request to Geocoding API failed with status code {response.status_code}."
            )
            logger.info("You might have to wait a while, before you can use it again)")
            logger.info("tip: Use automatic location setting to avoid this error. ")
        return None, None

    def _get_location(self) -> Dict:
        """
        Retrieves the geolocation of the current machine using the 'ipinfo.io' service.

        The method sends a GET request to 'ipinfo.io' which returns a JSON response
        with the IP, city, region, country, and geographic coordinates (latitude and longitude)
        of the machine from where the request was made.
        Returns:
            Dict: A dictionary containing the following keys:
                'ip': The IP address of the machine.
                'city': The city in which the machine is located.
                'region': The region in which the machine is located.
                'country': The country in which the machine is located.
                'loc': The latitude and longitude of the machine's location.
        """

        # Send GET request to 'ipinfo.io' to fetch geolocation data
        response = requests.get(conf.ipinfo.url)
        # Convert the response to JSON and return it
        return response.json()



#!/usr/bin/env python3

import argparse
import sys
import warnings
from argparse import RawTextHelpFormatter

from loguru import logger

from renops.config import OptimisationType
from renops.geoshifter import GeoShift
from renops.scheduler import Scheduler, execute_script
from renops.utils import read_json_from_filename

warnings.simplefilter("always", DeprecationWarning)


def main():
    try:
        run()

    except ValueError as error:
        logger.error(f"ValueError: {error}")
        sys.exit(1)  # Exiting with status code 1 signifies an error

    except KeyboardInterrupt:
        logger.error("\nKeyboard interrupt detected. Exiting.")
        sys.exit(0)  # Exiting with status code 0 signifies a clean exit

    except Exception as ex:
        logger.error(f"An unexpected error occurred: {ex}")
        sys.exit(1)  # Exiting with status code 1 for error


def run():
    logger.info("RUNNING RENOPS SCHEDULER...")
    parser = argparse.ArgumentParser(formatter_class=RawTextHelpFormatter)
    parser.add_argument(
        "script_path",
        help="Path to the script to be executed or JSON file in case of geo shifting.",
    )
    parser.add_argument(
        "-l",
        "--location",
        default=None,
        help=(
            "Location can be specified in two ways:\n\n"
            '1. Pass a specific location as a string, e.g., "Berlin, Germany".\n\n'
            "2. Use automatic location detection based on IP address.\n"
            " By using this tag, you agree that your IP can be used to detect your location.\n"
            "You can use any of the following values for this purpose:\n"
            "   -l a (-la)\n"
            "   -l auto\n"
            "   -l automatic\n"
        ),
    )

    parser.add_argument(
        "-gs",
        "--geo-shift",
        action="store_true",
        help="JSON on given path should be formated as:\n"
        "{\n"
        '  "hpc1": {\n'
        '    "location": "Berlin, Germany",\n'
        '    "cmd": "ssh user@hpc1 python3 train.py"\n'
        "  },\n"
        '  "hpc2": {\n'
        '    "location": "Madrid, Spain",\n'
        '    "cmd": "ssh user@hpc2 python3 train.py"\n'
        "  },\n"
        '  "hpc3": {\n'
        '    "location": "Copenhagen, Denmark",\n'
        '    "cmd": "ssh user@hpc3 python3 train.py"\n'
        "  }\n"
        "}",
    )

    parser.add_argument(
        "-r",
        "--runtime",
        type=int,
        default=None,
        help="Runtime in hours. (Not for geo shift mode)",
    )

    parser.add_argument(
        "-d",
        "--deadline",
        type=int,
        default=120,
        help="Deadline in hours, by when should script finish running (Not for geo shift mode)",  # noqa
    )

    parser.add_argument(
        "-o",
        "--optimise",
        choices=["renewable", "price", "emissions"],
        required=False,
        help=(
            "Choose an optimisation type:\n"
            " - 'renewable' (renewable potential - renewable energy availability on a scale from 0 to 1)\n"
            " - 'price' (day-ahead energy price)\n"
            " - 'emissions' (Carbon emissions in gCO2eq/kWh)\n"
        ),
    )
    parser.add_argument(
        "-op", "--optimise-price", action="store_true", help=argparse.SUPPRESS
    )

    parser.add_argument("-v", "--verbose", action="store_true", help="Verbose mode.")

    args = parser.parse_args()

    # Check for the deprecated argument and warn the user
    if args.optimise_price:
        warnings.warn(
            "'--optimise-price' is deprecated and will be removed in future versions. "
            "Use '--optimise price' instead.",
            DeprecationWarning,
        )
        # Map the deprecated argument to the new one if necessary
        args.optimise = "price"

    if args.optimise is None:
        warnings.warn(
            "Default optimisation type is deprecated and will be mandatory in future versions. "
            "Setting optimise flag to 'renewable'. "
            "Use '--optimise ' instead.",
            DeprecationWarning,
        )
        args.optimise = "renewable"

    optimisation_map = {
        "renewable": OptimisationType.renewable_potential,
        "price": OptimisationType.price,
        "emissions": OptimisationType.carbon_emissions,
    }

    optimise_type = optimisation_map[args.optimise].value

    if args.geo_shift:
        logger.info("Geo shift mode specified, shifting in space...")
        if not args.script_path.endswith(".json"):
            raise ValueError("The input file must be a JSON file.")
        s = GeoShift(
            locations=read_json_from_filename(args.script_path),
            optimise_type=optimise_type,
            verbose=args.verbose,
        )
        s.shift()

    elif args.location:
        logger.info("Location specified, shifting in time...")
        args = parser.parse_args()
        if not args.runtime:
            logger.info("Runtime not specified, using default setting of 3 hours!")
            args.runtime = 3

        s = Scheduler(
            deadline=args.deadline,
            runtime=args.runtime,
            location=args.location,
            optimise_type=optimise_type,
            verbose=args.verbose,
            action=execute_script,
            argument=([args.script_path]),
        )
        s.run()

    else:
        raise ValueError(
            "Specifiy either location (-l) or geo shift mode (-gs). Check --help for more details."
        )  # noqa


if __name__ == "__main__":
    main()



import json
import subprocess

import numpy as np
import pandas as pd


def get_closest_metric(
    data, current_epoch: int, metric_col="metric", epoch_col="epoch"
) -> float:
    """
    This function takes a pandas DataFrame, a current epoch value,
    and column names for metric and epoch as input.
    It returns the value of the "metric" column for the row that
      has the closest epoch to the current epoch.

    Args:
        data (pd.DataFrame): The pandas DataFrame containing the data.
        current_epoch (int): The current epoch value.
        metric_col (str, optional): The name of the column containing
                                    the metric values. Defaults to "metric".
        epoch_col (str, optional): The name of the column containing the epoch
                                 values. Defaults to "epoch".

    Returns:
        float: The value of the "metric" column for
               the row with the closest epoch.
    """

    # Ensure numerical data types for epoch
    data[epoch_col] = pd.to_numeric(data[epoch_col], errors="coerce")

    # Handle potential missing values in epoch column
    if data[epoch_col].isnull().any():
        raise ValueError("Epoch column contains missing values")

    # Calculate absolute differences between current epoch and all epochs
    epoch_diffs = np.abs(data[epoch_col] - current_epoch)

    # Find the index of the row with the closest epoch
    closest_epoch_idx = epoch_diffs.idxmin()

    # Return the corresponding metric value
    return data.loc[closest_epoch_idx, metric_col]


def execute_linux_command(command):
    process = subprocess.Popen(
        command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    output, errors = process.communicate()
    return output.decode(), errors.decode()


def read_json_from_filename(filename):
    with open(filename, "r") as file:
        data = json.load(file)
    return data
